@@ -196,15 +196,15 @@ r6 = room {
 	end,
 	dsc = [[Lección 6. Moviéndose - Parte II^^
 		Ahora ve a la última lección.]],
-	way = { 'r1', 'r2', 'r3', 'r4', 'r6', 'theend'},
+	way = { 'r1', 'r2', 'r3', 'r4', 'r5', 'theend'},
 };
 
 theend = room {
 	nam = 'Ultima lección',
 	dsc = [[Mediante el menú flotante, puedes elegir un juego, salvar o cargar un juego, y un par de cosas más. Para abrirlo presiona "Esc" o haz click en el icono de menú abajo a la derecha.^^
 		Ahora ya estás listo para jugar. Buena suerte!!!^^
-		Descarga nuevos juegos para INSTEAD aquí: ]]..txtu("http://instead.sourceforge.net")..[[^^
-		Este tutorial usa el tema Instead game-engine de svenzzon.]],
+		Descarga nuevos juegos para INSTEAD aquí: ]]..txtu("http://instead-games.sourceforge.net")..[[^^
+		Este tutorial usa el tema "Instead game-engine" de svenzzon.]],
 	obj = { vway('keys', 'Ver {lista de atajos de teclado}.', 'help')},
 };
 
@@ -243,7 +243,6 @@ help2 = room {
 	nam = 'Parámetros línea de comandos',
 	dsc = [[
 	-nosound - lanza sin sonido;^
-	-alsa - fuerza el uso de ALSA (solo Linux);^
 	-debug - modo debug para desarrolladores;^
 	-theme <theme> - elige un tema;^
 	-game <game> - elige un juego;^
