@@ -824,6 +824,8 @@ static int instead_platform(void)
 	snprintf(plat, sizeof(plat) - 1, "PLATFORM='MACOSX'");
 #elif defined(_WIN32_WCE)
 	snprintf(plat, sizeof(plat) - 1, "PLATFORM='WINCE'");
+#elif defined(WINRT)
+	snprintf(plat, sizeof(plat) - 1, "PLATFORM='WINRT'");
 #elif defined(S60)
 	snprintf(plat, sizeof(plat) - 1, "PLATFORM='S60'");
 #elif defined(ANDROID)
@@ -832,6 +834,8 @@ static int instead_platform(void)
 	snprintf(plat, sizeof(plat) - 1, "PLATFORM='WIN32'");
 #elif defined(MAEMO)
 	snprintf(plat, sizeof(plat) - 1, "PLATFORM='MAEMO'");
+#elif defined(SAILFISHOS)
+	snprintf(plat, sizeof(plat) - 1, "PLATFORM='SFOS'");
 #else
 	snprintf(plat, sizeof(plat) - 1, "PLATFORM='UNIX'");
 #endif
@@ -848,7 +852,7 @@ static int instead_package(const char *path)
 	if (!stead_path)
 		return -1;
 	strcpy(stead_path, "package.path=\"");
-#if defined(_WIN32_WCE)
+#if defined(_WIN32_WCE) || defined(WINRT)
 	if (path) {
 		strcat(stead_path, path); /* wince have not cwd :) */
 		strcat(stead_path, "/?.lua;");
@@ -897,7 +901,6 @@ static int instead_set_api(const char *api)
 	char *oa;
 	if (!api || !*api) {
 		FREE(API);
-		API = NULL;
 		snprintf(instead_api_path, sizeof(instead_api_path), "%s", STEAD_PATH);
 	} else {
 		s = strlen(api);
@@ -966,9 +969,15 @@ int instead_init_lua(const char *path, int detect)
 {
 	int api = 0;
 	busy = 0;
-	setlocale(LC_ALL,"");
-	setlocale(LC_NUMERIC,"C"); /* to avoid . -> , in numbers */	
-	setlocale(LC_CTYPE,"C"); /* to avoid lower/upper problems */	
+	setlocale(LC_ALL, "");
+	setlocale(LC_NUMERIC, "C"); /* to avoid . -> , in numbers */
+	setlocale(LC_CTYPE, "C"); /* to avoid lower/upper problems */
+#ifdef LC_MESSAGES
+	setlocale(LC_MESSAGES, "C");
+#endif
+#ifdef LC_COLLATE
+	setlocale(LC_COLLATE, "C");
+#endif
 /*	strcpy(curcp, "UTF-8"); */
 	getdir(instead_cwd_path, sizeof(instead_cwd_path));
 	unix_path(instead_cwd_path);
@@ -1109,22 +1118,17 @@ void instead_done(void)
 #endif
 #endif
 #ifdef _HAVE_ICONV
-	if (fromcp)
-		free(fromcp);
+	FREE(fromcp);
 #endif
 	if (L)
 		lua_close(L);
 	L = NULL;
-#ifdef _HAVE_ICONV
-	fromcp = NULL;
-#endif
 	if (data_idf)
 		idf_done(data_idf);
 	data_idf = NULL;
 	if (wasL)
 		setdir(instead_cwd_path);
 	FREE(API);
-	API = NULL;
 }
 
 int  instead_encode(const char *s, const char *d)
