@@ -1,5 +1,5 @@
 /*
- * Copyright 2009-2017 Peter Kosyh <p.kosyh at gmail.com>
+ * Copyright 2009-2018 Peter Kosyh <p.kosyh at gmail.com>
  *
  * Permission is hereby granted, free of charge, to any person
  * obtaining a copy of this software and associated documentation files
@@ -276,6 +276,7 @@ static int luaB_text_size(lua_State *L) {
 
 	const char *font = luaL_optstring(L, 1, NULL);
 	const char *text = luaL_optstring(L, 2, NULL);
+	int style = luaL_optnumber(L, 3, 0);
 
 	if (!font)
 		return 0;
@@ -291,6 +292,7 @@ static int luaB_text_size(lua_State *L) {
 		else
 			h = fnt_height(fn->fnt);
 	} else {
+		fnt_style(fn->fnt, style);
 		txt_size(fn->fnt, text, &w, &h);
 		if (fn->flags & FN_SCALED) {
 			w = ceil((float)w / game_theme.scale);
@@ -366,13 +368,28 @@ err:
 	return 0;
 }
 
+static int in_callback = 0;
+
 static img_t grab_sprite(const char *dst, int *xoff, int *yoff)
 {
+	img_t oldscreen;
 	img_t d;
-	if (DIRECT_MODE && dst && !strcmp(dst, "screen")) {
+	if (!dst)
+		return NULL;
+	if ((DIRECT_MODE || in_callback) && !strcmp(dst, "screen")) {
 		d = gfx_screen(NULL);
 		*xoff = game_theme.xoff;
 		*yoff = game_theme.yoff;
+	} else if (opt_owntheme && !strcmp(dst, "screen")) {
+		if (!game_theme.bg) { /* create on the fly */
+			game_theme.bg = gfx_new(game_theme.w, game_theme.h);
+			oldscreen = gfx_screen(game_theme.bg);
+			gfx_clear(0, 0, game_theme.w, game_theme.h);
+			gfx_screen(oldscreen);
+		}
+		d = game_theme.bg;
+		*xoff = game_theme.xoff;
+		*yoff = game_theme.yoff;
 	} else {
 		*xoff = 0;
 		*yoff = 0;
@@ -410,7 +427,6 @@ static int luaB_sprite_size(lua_State *L) {
 
 static int luaB_blit_sprite(lua_State *L, int mode) {
 	img_t s = NULL, d = NULL;
-	img_t img2 = NULL;
 	float v;
 	struct lua_pixels *pixels = lua_touserdata(L, 1);
 	const char *src = NULL;
@@ -462,16 +478,14 @@ static int luaB_blit_sprite(lua_State *L, int mode) {
 
 	game_pict_modify(d);
 
-	if (alpha != 255) {
-		img2 = gfx_alpha_img(s, alpha);
-		if (img2)
-			s = img2;
-	}
 	game_gfx_clip();
 
 	switch (mode) {
 	case BLIT_DRAW:
-		gfx_draw_from(s, x + xoff0, y + yoff0, w, h, d, xx + xoff, yy + yoff);
+		if (alpha != 255)
+			gfx_draw_from_alpha(s, x + xoff0, y + yoff0, w, h, d, xx + xoff, yy + yoff, alpha);
+		else
+			gfx_draw_from(s, x + xoff0, y + yoff0, w, h, d, xx + xoff, yy + yoff);
 		break;
 	case BLIT_COPY:
 		gfx_copy_from(s, x + xoff0, y + yoff0, w, h, d, xx + xoff, yy + yoff);
@@ -483,8 +497,7 @@ static int luaB_blit_sprite(lua_State *L, int mode) {
 		break;
 	}
 
-	gfx_noclip();
-	gfx_free_image(img2);
+	game_gfx_noclip();
 	lua_pushboolean(L, 1);
 	return 1;
 }
@@ -728,7 +741,7 @@ static int luaB_fill_sprite(lua_State *L) {
 	game_pict_modify(d);
 	game_gfx_clip();
 	gfx_img_fill(d, x + xoff, y + yoff, w, h, col);
-	gfx_noclip();
+	game_gfx_noclip();
 	lua_pushboolean(L, 1);
 	return 1;
 }
@@ -855,7 +868,7 @@ static int luaB_theme_var(lua_State *L) {
 		return 0;
 	if (!theme_setvar((char*)var, (char*)val)) {
 		if (strcmp(var, "win.scroll.mode")) /* let change scroll mode w/o theme reload */
-			game_theme_changed = 2;
+			game_theme_changed = 1;
 	}
 	return 0;
 }
@@ -878,13 +891,55 @@ static int luaB_theme_name(lua_State *L) {
 	return 1;
 }
 
+static int luaB_instead_direct(lua_State *L) {
+	int direct = -1;
+	int old = DIRECT_MODE;
+
+	if (lua_isboolean(L, 1))
+		direct = lua_toboolean(L, 1);
+
+	if (direct == -1) {
+		lua_pushboolean(L, old);
+		return 1;
+	}
+
+	if (!opt_owntheme) {
+		lua_pushboolean(L, 0);
+		return 1;
+	}
+
+	if (direct)
+		game_theme.gfx_mode = GFX_DIRECT_SET(game_theme.gfx_mode);
+	else {
+		if (game_theme.gfx_mode != GFX_MODE_DIRECT) {
+			game_theme.gfx_mode = GFX_DIRECT_CLR(game_theme.gfx_mode);
+		}
+	}
+	lua_pushboolean(L, 1);
+	return 1;
+}
+static unsigned long busy_time = 0;
+
+void instead_ready(void)
+{
+	if (menu_visible() == menu_wait) {
+		menu_toggle(-1);
+	}
+	busy_time = 0;
+}
 static int luaB_stead_busy(lua_State *L) {
-	static unsigned long busy_time = 0;
 	int busy = lua_toboolean(L, 1);
 	if (busy) {
 		struct inp_event ev;
 		int dirty = 0;
 		memset(&ev, 0, sizeof(ev));
+
+		if (!game_freezed()) {
+			if (game_bg_modify(NULL))
+				game_redraw_all();
+			game_flip();
+		}
+
 		while (input(&ev, 0) == AGAIN);
 		if (ev.type == MOUSE_MOTION) {
 			game_cursor(CURSOR_ON); /* to make all happy */
@@ -900,10 +955,7 @@ static int luaB_stead_busy(lua_State *L) {
 			game_gfx_commit(0);
 		return 0;
 	}
-	if (menu_visible() == menu_wait) {
-		menu_toggle(-1);
-	}
-	busy_time = 0;
+	instead_ready();
 	return 0;
 }
 
@@ -1007,7 +1059,7 @@ static int pixels_size(lua_State *L) {
 
 #define PXL_BLEND_COPY 1
 #define PXL_BLEND_BLEND 2
-static inline void blend(unsigned char *s, unsigned char *d)
+static __inline void blend(unsigned char *s, unsigned char *d)
 {
 	unsigned int r, g, b, a;
 	unsigned int sa = s[3];
@@ -1022,7 +1074,7 @@ static inline void blend(unsigned char *s, unsigned char *d)
 	d[0] = r; d[1] = g; d[2] = b; d[3] = a;
 }
 
-static inline void draw(unsigned char *s, unsigned char *d)
+static __inline void draw(unsigned char *s, unsigned char *d)
 {
 	unsigned int r, g, b, a;
 	unsigned int sa = s[3];
@@ -1035,7 +1087,7 @@ static inline void draw(unsigned char *s, unsigned char *d)
 		((unsigned int)d[2] * (255 - sa) >> 8);
 	d[0] = r; d[1] = g; d[2] = b; d[3] = a;
 }
-static inline void pixel(unsigned char *s, unsigned char *d)
+static __inline void pixel(unsigned char *s, unsigned char *d)
 {
 	unsigned char a_src = s[3];
 	unsigned char a_dst = d[3];
@@ -1049,7 +1101,7 @@ static inline void pixel(unsigned char *s, unsigned char *d)
 		blend(s, d);
 	}
 }
-static inline void line0(struct lua_pixels *hdr, int x1, int y1, int dx, int dy, int xd, unsigned char *col)
+static __inline void line0(struct lua_pixels *hdr, int x1, int y1, int dx, int dy, int xd, unsigned char *col)
 {
 	int dy2 = dy * 2;
 	int dyx2 = dy2 - dx * 2;
@@ -1094,7 +1146,7 @@ static inline void line0(struct lua_pixels *hdr, int x1, int y1, int dx, int dy,
 	return;
 }
 
-static inline void line1(struct lua_pixels *hdr, int x1, int y1, int dx, int dy, int xd, unsigned char *col)
+static __inline void line1(struct lua_pixels *hdr, int x1, int y1, int dx, int dy, int xd, unsigned char *col)
 {
 	int dx2 = dx * 2;
 	int dxy2 = dx2 - dy * 2;
@@ -1436,12 +1488,12 @@ static void _fill(struct lua_pixels *src, int x, int y, int w, int h,
 	return;
 }
 
-static inline int orient2d(int ax, int ay, int bx, int by, int cx, int cy)
+static __inline int orient2d(int ax, int ay, int bx, int by, int cx, int cy)
 {
 	return (bx - ax) * (cy - ay) - (by - ay) * (cx - ax);
 }
 
-static inline int min3(int a, int b, int c)
+static __inline int min3(int a, int b, int c)
 {
 	if (a < b) {
 		if (a < c)
@@ -1454,7 +1506,7 @@ static inline int min3(int a, int b, int c)
 	}
 }
 
-static inline int max3(int a, int b, int c)
+static __inline int max3(int a, int b, int c)
 {
 	if (a > b) {
 		if (a > c)
@@ -2398,6 +2450,55 @@ static int luaB_noise4(lua_State *L) {
 	return 1;
 }
 
+static int callback_ref = 0;
+static int render_callback_dirty = 0;
+
+int instead_render_callback_dirty(int fl)
+{
+	int rc = render_callback_dirty;
+	if (!callback_ref || game_freezed())
+		return 0;
+	if (fl != -1)
+		render_callback_dirty = fl;
+	return rc;
+}
+
+void instead_render_callback(void)
+{
+	if (!callback_ref || game_freezed() || render_callback_dirty == -1)
+		return;
+
+	game_cursor(CURSOR_CLEAR);
+	instead_lock();
+	lua_rawgeti(instead_lua(), LUA_REGISTRYINDEX, callback_ref);
+	in_callback ++;
+	if (instead_pcall(instead_lua(), 0)) { /* on any error */
+		luaL_unref(instead_lua(), LUA_REGISTRYINDEX, callback_ref);
+		callback_ref = 0;
+	}
+	in_callback --;
+	instead_clear();
+	instead_unlock();
+	if (game_pict_modify(NULL))
+		render_callback_dirty = -1;
+	game_cursor(CURSOR_DRAW);
+	return;
+}
+
+static int luaB_after_callback(lua_State *L) {
+	if (!opt_owntheme) {
+		lua_pushboolean(L, 0);
+		return 1;
+	}
+	if (callback_ref)
+		luaL_unref(L, LUA_REGISTRYINDEX, callback_ref);
+	callback_ref = 0;
+	if (lua_isfunction(L, 1))
+		callback_ref = luaL_ref(L, LUA_REGISTRYINDEX);
+	lua_pushboolean(L, 0);
+	return 1;
+}
+
 static const luaL_Reg sprites_funcs[] = {
 	{"instead_font_load", luaB_load_font},
 	{"instead_font_free", luaB_free_font},
@@ -2423,6 +2524,7 @@ static const luaL_Reg sprites_funcs[] = {
 	{"instead_theme_name", luaB_theme_name},
 	{"instead_ticks", luaB_get_ticks},
 	{"instead_busy", luaB_stead_busy},
+	{"instead_direct", luaB_instead_direct},
 	{"instead_mouse_pos", luaB_mouse_pos},
 	{"instead_mouse_filter", luaB_mouse_filter},
 	{"instead_mouse_show", luaB_mouse_show},
@@ -2432,11 +2534,17 @@ static const luaL_Reg sprites_funcs[] = {
 	{"instead_noise2", luaB_noise2},
 	{"instead_noise3", luaB_noise3},
 	{"instead_noise4", luaB_noise4},
+	{"instead_render_callback", luaB_after_callback},
 	{NULL, NULL}
 };
 
 static int sprites_done(void)
 {
+	if (callback_ref) {
+		luaL_unref(instead_lua(), LUA_REGISTRYINDEX, callback_ref);
+		callback_ref = 0;
+		render_callback_dirty = 0;
+	}
 	sprites_free();
 	return 0;
 }
@@ -2451,9 +2559,19 @@ static int sprites_init(void)
 	return instead_loadfile(dirpath(path));
 }
 
+static int sprites_err(void)
+{
+	if (callback_ref) {
+		luaL_unref(instead_lua(), LUA_REGISTRYINDEX, callback_ref);
+		callback_ref = 0;
+	}
+	return 0;
+}
+
 static struct instead_ext ext = {
 	.init = sprites_init,
 	.done = sprites_done,
+	.err = sprites_err,
 };
 
 int instead_sprites_init(void)
