@@ -1,5 +1,5 @@
 /*
- * Copyright 2009-2017 Peter Kosyh <p.kosyh at gmail.com>
+ * Copyright 2009-2018 Peter Kosyh <p.kosyh at gmail.com>
  *
  * Permission is hereby granted, free of charge, to any person
  * obtaining a copy of this software and associated documentation files
@@ -22,7 +22,7 @@
  *
  */
 
-#if defined(__APPLE__) || defined(S60) || defined(ANDROID)
+#if defined(__APPLE__) || defined(S60) || defined(ANDROID) || defined(WINRT)
 #include <SDL.h>
 #endif
 
@@ -60,6 +60,7 @@ int vsync_sw = 0;
 int resizable_sw = 0;
 int scale_sw = 1;
 int standalone_sw = 0;
+int nocursor_sw = 0;
 
 static int opt_index = 1;
 
@@ -78,6 +79,7 @@ char *appdata_sw = NULL;
 char *idf_sw = NULL;
 char *start_idf_sw = NULL;
 char *lua_sw = NULL;
+char *render_sw = NULL;
 static int lua_exec = 1;
 static int nocfg_sw = 0;
 
@@ -209,7 +211,7 @@ static int run_game(const char *path)
 }
 static void usage(void)
 {
-	fprintf(stderr, 
+	fprintf(stderr,
 	"INSTEAD "VERSION" - Simple Text Adventure Engine, The Interpreter\n"
 	"Usage:\n"
 	"    sdl-instead [options] [game.zip or game.idf or path to game]\n"
@@ -235,9 +237,64 @@ static void usage(void)
 static int profile_load(const char *path);
 extern int game_instead_extensions(void);
 
+static int luaB_clipboard(lua_State *L) {
+	char *buf = NULL;
+	const char *text = luaL_optstring(L, 1, NULL);
+	if (!text) { /* get */
+		if (system_clipboard(NULL, &buf) == 0) {
+			lua_pushstring(L, buf);
+			free(buf);
+		} else
+			lua_pushboolean(L, 0);
+		return 1;
+	}
+	lua_pushboolean(L, system_clipboard(text, NULL) == 0);
+	return 1;
+}
+
+static int luaB_wait_use(lua_State *L) {
+	int v = -1;
+	int old = game_wait_use;
+
+	if (lua_isboolean(L, 1))
+		v = lua_toboolean(L, 1);
+
+	if (v == -1) {
+		lua_pushboolean(L, old);
+		return 1;
+	}
+
+	if (!opt_owntheme) {
+		lua_pushboolean(L, 0);
+		return 1;
+	}
+	game_wait_use = v;
+	lua_pushboolean(L, old);
+	return 1;
+}
+
+static int luaB_grab_events(lua_State *L) {
+	int grab = lua_toboolean(L, 1);
+	int ov = game_grab_events;
+	if (lua_isboolean(L, 1))
+		game_grab_events = grab;
+	lua_pushboolean(L, ov);
+	return 1;
+}
+
+static const luaL_Reg sdl_funcs[] = {
+	{ "instead_clipboard", luaB_clipboard },
+	{ "instead_wait_use", luaB_wait_use },
+	{"instead_grab_events", luaB_grab_events},
+	{NULL, NULL}
+};
+
 static int sdl_ext_init(void)
 {
 	char path[PATH_MAX];
+	instead_api_register(sdl_funcs);
+	game_wait_use = 1;
+	game_grab_events = 0;
 	snprintf(path, sizeof(path), "%s/%s", instead_stead_path(), "/ext/gui.lua");
 	return instead_loadfile(dirpath(path));
 }
@@ -270,18 +327,17 @@ int instead_main(int argc, char *argv[])
 #ifdef _WIN32_WCE
 	libwince_init(argv[0], 1);
 	wince_init(argv[0]);
-#else
-#ifdef S60
+#elif defined(WINRT)
+	unix_path(argv[0]);
+	strcpy(game_cwd, argv[0]);
+#elif defined(S60)
 	extern char s60_data[];
 	strcpy(game_cwd, s60_data);
-#else
-#ifdef _WIN32
+#elif defined(_WIN32)
 	strcpy(game_cwd, dirname(argv[0]));
 #else
 	if (!getcwd(game_cwd, sizeof(game_cwd)))
 		fprintf(stderr,"Warning: can not get current dir\n.");
-#endif
-#endif
 #endif
 	if (sdl_extensions() < 0) {
 		fprintf(stderr, "Fatal: can not init SDL extensions\n");
@@ -400,6 +456,8 @@ int instead_main(int argc, char *argv[])
 			version_sw = 1;
 		} else if (!strcmp(argv[i], "-nopause")) {
 			nopause_sw = 1;
+		} else if (!strcmp(argv[i], "-nocursor")) {
+			nocursor_sw = 1;
 		} else if (!strcmp(argv[i], "-software")) {
 			software_sw = 1;
 		} else if (!strcmp(argv[i], "-resizable")) {
@@ -457,6 +515,16 @@ int instead_main(int argc, char *argv[])
 			|| !strcmp(argv[i], "--help")) {
 			usage();
 			goto out;
+		} else if (!strcmp(argv[i], "-renderer")) {
+			FREE(render_sw);
+			if ((i + 1) < argc)
+				render_sw = strdup(argv[++i]);
+			else {
+				fprintf(stderr, "Parameter required: %s\n", argv[i]);
+				fprintf(stderr, "opengl, opengles2, opengles, software, direct3d\n");
+				err = 1;
+				goto out;
+			}
 		} else if (argv[i][0] == '-') {
 			fprintf(stderr,"Unknown option: %s\n", argv[i]);
 			usage();
@@ -556,7 +624,7 @@ int instead_main(int argc, char *argv[])
 		goto out;
 	}
 	menu_langs_lookup(dirpath(LANG_PATH));
-	
+
 	if (!langs_nr) {
 		fprintf(stderr, "No languages found in: %s.\n", dirpath(LANG_PATH));
 		err = 1;
@@ -564,13 +632,13 @@ int instead_main(int argc, char *argv[])
 	}
 	if (!opt_lang || !opt_lang[0])
 		opt_lang = game_locale();
-	
+
 	if (menu_lang_select(opt_lang) && menu_lang_select(LANG_DEF)) {
 		fprintf(stderr, "Can not load default language.\n");
 		err = 1;
 		goto out;
 	}
-	
+
 	if (games_sw)
 		games_lookup(games_sw);
 
@@ -606,7 +674,7 @@ int instead_main(int argc, char *argv[])
 			}
 		}
 		if (d)
-			free(d); 
+			free(d);
 		if (b)
 			free(b);
 	}
@@ -641,25 +709,28 @@ int instead_main(int argc, char *argv[])
 
 	if (mode_sw)
 		parse_mode(mode_sw, opt_mode);
-	
+
 	if (game_sw) {
 		FREE(opt_game);
 		opt_game = game_sw;
-	}	
+	}
 
 	if (theme_sw) {
 		FREE(opt_theme);
 		opt_theme = theme_sw;
 	}
-	
+
 	if (opt_theme)
 		game_theme_select(opt_theme);
 	if (!curtheme_dir[THEME_GLOBAL])
 		game_theme_select(DEFAULT_THEME);
-	
+
 	/* Initialize SDL */
 	if (gfx_init() < 0)
 		return -1;
+	/* Initialize Sound */
+	snd_init(opt_hz);
+	snd_volume_mus(opt_vol);
 #ifdef _USE_GTK
 	gtk_init(&argc, &argv); /* must be called AFTER SDL_Init when using SDL2 */
 #endif
@@ -677,8 +748,8 @@ int instead_main(int argc, char *argv[])
 	cfg_save();
 	game_done(0);
 
+	snd_done();
 	gfx_video_done();
-
 #ifndef ANDROID
 	gfx_done();
 #endif
@@ -722,6 +793,8 @@ static struct parser profile_parser[] = {
 	{ "mode", parse_string, &mode_sw, 0 },
 	{ "modes", parse_string, &modes_sw, 0 },
 	{ "fontscale", parse_string, &fsize_sw, 0 },
+	{ "renderer", parse_string, &render_sw, 0 },
+	{ "nocursor", parse_int, &nocursor_sw, 0 },
 	{ NULL, NULL, NULL, 0 },
 };
 
