@@ -1,11 +1,7 @@
 #ifndef S60_H
 #define S60_H
 
-#ifndef PATH_MAX
-#define PATH_MAX 256
-#endif
-
-#define VERSION "3.1.2"
+#define VERSION "3.2.1"
 #define DATAPATH "./"
 #define STEAD_PATH DATAPATH"stead/"
 #define THEMES_PATH DATAPATH"themes/"
