@@ -1,5 +1,5 @@
 /*
- * Copyright 2009-2017 Peter Kosyh <p.kosyh at gmail.com>
+ * Copyright 2009-2018 Peter Kosyh <p.kosyh at gmail.com>
  *
  * Permission is hereby granted, free of charge, to any person
  * obtaining a copy of this software and associated documentation files
@@ -50,7 +50,7 @@ static int parse_gfx_mode(const char *v, void *data)
 {
 	int *i = (int *)data;
 	if (!strcmp(v, "fixed"))
-		*i = GFX_MODE_FIXED;	
+		*i = GFX_MODE_FIXED;
 	else if (!strcmp(v, "embedded"))
 		*i = GFX_MODE_EMBEDDED;
 	else if (!strncmp(v, "float", 5)) {
@@ -84,10 +84,10 @@ static int parse_gfx_mode(const char *v, void *data)
 				return -1;
 		}
 	} else if (!strcmp(v, "direct"))
-		*i = GFX_MODE_DIRECT;	
+		*i = GFX_MODE_DIRECT;
 	else
 		return -1;
-	return 0;	
+	return 0;
 }
 
 static int out_gfx_mode(const void *v, char **out)
@@ -251,7 +251,7 @@ static int theme_parse_full_path(const char *v, void *data)
 	char **p = (char **)data;
 	char *np;
 	if (theme_relative) {
-		if (!strncmp(v, "blank:", 6) || 
+		if (!strncmp(v, "blank:", 6) ||
 			!strncmp(v, "box:", 4) ||
 			!strncmp(v, "spr:", 4)) /* hack for special files*/
 			return parse_path(v, data);
@@ -273,7 +273,7 @@ static int theme_parse_full_path(const char *v, void *data)
 struct parser cmd_parser[] = {
 	{ "scr.w", parse_int, &game_theme.w, 0 },
 	{ "scr.h", parse_int, &game_theme.h, 0 },
-	{ "scr.gfx.scalable", parse_int, &game_theme.gfx_scalable, 0 },
+	{ "scr.gfx.scalable", parse_int, &game_theme.gfx_scalable, CHANGED_WIN | CHANGED_INV | CHANGED_FONT | CHANGED_IFONT | CHANGED_MFONT },
 	{ "scr.col.bg", parse_color, &game_theme.bgcol, 0 },
 	{ "scr.gfx.icon", theme_parse_full_path, &game_theme.icon_name, CHANGED_ICON },
 	{ "scr.gfx.bg", theme_parse_full_path, &game_theme.bg_name, CHANGED_BG },
@@ -282,56 +282,56 @@ struct parser cmd_parser[] = {
 	{ "scr.gfx.cursor.y", parse_int, &game_theme.cur_y, 0 },
 	{ "scr.gfx.use", theme_parse_full_path, &game_theme.use_name, CHANGED_USE }, /* compat */
 	{ "scr.gfx.cursor.use", theme_parse_full_path, &game_theme.use_name, CHANGED_USE },
-	{ "scr.gfx.pad", parse_int, &game_theme.pad, 0 }, 
+	{ "scr.gfx.pad", parse_int, &game_theme.pad, CHANGED_WIN | CHANGED_INV },
 	{ "scr.gfx.x", parse_int, &game_theme.gfx_x, 0 },
 	{ "scr.gfx.y", parse_int, &game_theme.gfx_y, 0 },
 	{ "scr.gfx.w", parse_int, &game_theme.max_scene_w, 0 },
 	{ "scr.gfx.h", parse_int, &game_theme.max_scene_h, 0 },
-	{ "scr.gfx.mode", parse_gfx_mode, &game_theme.gfx_mode, 0 },
+	{ "scr.gfx.mode", parse_gfx_mode, &game_theme.gfx_mode, CHANGED_WIN },
 
-	{ "win.align", parse_win_align, &game_theme.win_align, 0 },
-	{ "win.x", parse_int, &game_theme.win_x, 0 },
-	{ "win.y", parse_int, &game_theme.win_y, 0 },
-	{ "win.w", parse_int, &game_theme.win_w, 0 },
-	{ "win.h", parse_int, &game_theme.win_h, 0 },
-	{ "win.ways.mode", parse_ways_mode, &game_theme.ways_mode, 0 },
-	{ "win.scroll.mode", parse_int, &game_theme.win_scroll_mode, 0 },
+	{ "win.align", parse_win_align, &game_theme.win_align, CHANGED_WIN },
+	{ "win.x", parse_int, &game_theme.win_x, CHANGED_WIN },
+	{ "win.y", parse_int, &game_theme.win_y, CHANGED_WIN },
+	{ "win.w", parse_int, &game_theme.win_w, CHANGED_WIN },
+	{ "win.h", parse_int, &game_theme.win_h, CHANGED_WIN },
+	{ "win.ways.mode", parse_ways_mode, &game_theme.ways_mode, CHANGED_WIN },
+	{ "win.scroll.mode", parse_int, &game_theme.win_scroll_mode, CHANGED_WIN },
 
-	{ "win.fnt.name", theme_parse_full_path, &game_theme.font_name, CHANGED_FONT },
-	{ "win.fnt.size", parse_int, &game_theme.font_size, CHANGED_FONT },
-	{ "win.fnt.height", parse_float, &game_theme.font_height, 0 },
+	{ "win.fnt.name", theme_parse_full_path, &game_theme.font_name, CHANGED_FONT | CHANGED_WIN },
+	{ "win.fnt.size", parse_int, &game_theme.font_size, CHANGED_FONT | CHANGED_WIN },
+	{ "win.fnt.height", parse_float, &game_theme.font_height, CHANGED_WIN },
 /* compat mode directive */
-	{ "win.gfx.h", parse_int, &game_theme.max_scene_h, 0 },
+	{ "win.gfx.h", parse_int, &game_theme.max_scene_h, CHANGED_WIN },
 /* here it was */
-	{ "win.gfx.up", theme_parse_full_path, &game_theme.a_up_name, CHANGED_UP },
-	{ "win.gfx.down", theme_parse_full_path, &game_theme.a_down_name, CHANGED_DOWN },
-	{ "win.up.x", parse_int, &game_theme.a_up_x, 0 },
-	{ "win.up.y", parse_int, &game_theme.a_up_y, 0 },
-	{ "win.down.x", parse_int, &game_theme.a_down_x, 0 },
-	{ "win.down.y", parse_int, &game_theme.a_down_y, 0 },
-	{ "win.col.fg", parse_color, &game_theme.fgcol, 0 }, 
-	{ "win.col.link", parse_color, &game_theme.lcol, 0 },
-	{ "win.col.alink", parse_color, &game_theme.acol, 0 }, 
+	{ "win.gfx.up", theme_parse_full_path, &game_theme.a_up_name, CHANGED_UP | CHANGED_WIN},
+	{ "win.gfx.down", theme_parse_full_path, &game_theme.a_down_name, CHANGED_DOWN | CHANGED_WIN},
+	{ "win.up.x", parse_int, &game_theme.a_up_x, CHANGED_WIN },
+	{ "win.up.y", parse_int, &game_theme.a_up_y, CHANGED_WIN },
+	{ "win.down.x", parse_int, &game_theme.a_down_x, CHANGED_WIN },
+	{ "win.down.y", parse_int, &game_theme.a_down_y, CHANGED_WIN },
+	{ "win.col.fg", parse_color, &game_theme.fgcol, CHANGED_WIN },
+	{ "win.col.link", parse_color, &game_theme.lcol, CHANGED_WIN },
+	{ "win.col.alink", parse_color, &game_theme.acol, CHANGED_WIN },
 
-	{ "inv.x", parse_int, &game_theme.inv_x, 0 },
-	{ "inv.y", parse_int, &game_theme.inv_y, 0 },
-	{ "inv.w", parse_int, &game_theme.inv_w, 0 },
-	{ "inv.h", parse_int, &game_theme.inv_h, 0 },
-	{ "inv.mode", parse_inv_mode, &game_theme.inv_mode, 0 },
-	{ "inv.horiz", parse_inv_mode, &game_theme.inv_mode, 0 },
+	{ "inv.x", parse_int, &game_theme.inv_x, CHANGED_INV },
+	{ "inv.y", parse_int, &game_theme.inv_y, CHANGED_INV },
+	{ "inv.w", parse_int, &game_theme.inv_w, CHANGED_INV },
+	{ "inv.h", parse_int, &game_theme.inv_h, CHANGED_INV },
+	{ "inv.mode", parse_inv_mode, &game_theme.inv_mode, CHANGED_INV },
+	{ "inv.horiz", parse_inv_mode, &game_theme.inv_mode, CHANGED_INV },
 
-	{ "inv.col.fg", parse_color, &game_theme.icol, 0 },
-	{ "inv.col.link", parse_color, &game_theme.ilcol, 0 },
-	{ "inv.col.alink", parse_color, &game_theme.iacol, 0 }, 
-	{ "inv.fnt.name", theme_parse_full_path, &game_theme.inv_font_name, CHANGED_IFONT },
-	{ "inv.fnt.size", parse_int, &game_theme.inv_font_size, CHANGED_IFONT },
-	{ "inv.fnt.height", parse_float, &game_theme.inv_font_height, 0 },
-	{ "inv.gfx.up", theme_parse_full_path, &game_theme.inv_a_up_name, CHANGED_IUP },
-	{ "inv.gfx.down", theme_parse_full_path, &game_theme.inv_a_down_name, CHANGED_IDOWN },
-	{ "inv.up.x", parse_int, &game_theme.inv_a_up_x, 0 },
-	{ "inv.up.y", parse_int, &game_theme.inv_a_up_y, 0 },
-	{ "inv.down.x", parse_int, &game_theme.inv_a_down_x, 0 },
-	{ "inv.down.y", parse_int, &game_theme.inv_a_down_y, 0 },
+	{ "inv.col.fg", parse_color, &game_theme.icol, CHANGED_INV },
+	{ "inv.col.link", parse_color, &game_theme.ilcol, CHANGED_INV },
+	{ "inv.col.alink", parse_color, &game_theme.iacol, CHANGED_INV },
+	{ "inv.fnt.name", theme_parse_full_path, &game_theme.inv_font_name, CHANGED_IFONT | CHANGED_INV },
+	{ "inv.fnt.size", parse_int, &game_theme.inv_font_size, CHANGED_IFONT | CHANGED_INV },
+	{ "inv.fnt.height", parse_float, &game_theme.inv_font_height, CHANGED_INV },
+	{ "inv.gfx.up", theme_parse_full_path, &game_theme.inv_a_up_name, CHANGED_IUP | CHANGED_INV },
+	{ "inv.gfx.down", theme_parse_full_path, &game_theme.inv_a_down_name, CHANGED_IDOWN | CHANGED_INV},
+	{ "inv.up.x", parse_int, &game_theme.inv_a_up_x, CHANGED_INV },
+	{ "inv.up.y", parse_int, &game_theme.inv_a_up_y, CHANGED_INV },
+	{ "inv.down.x", parse_int, &game_theme.inv_a_down_x, CHANGED_INV },
+	{ "inv.down.y", parse_int, &game_theme.inv_a_down_y, CHANGED_INV },
 
 	{ "menu.col.bg", parse_color, &game_theme.menu_bg, 0 },
 	{ "menu.col.fg", parse_color, &game_theme.menu_fg, 0 },
@@ -370,7 +370,7 @@ static theme_scalable_t theme_scalables[] = {
 	{ "scr.h", &game_theme.h, 0 },
 	{ "scr.gfx.cursor.x", &game_theme.cur_x, 0 },
 	{ "scr.gfx.cursor.y", &game_theme.cur_y, 0 },
-	{ "scr.gfx.pad", &game_theme.pad, 0 }, 
+	{ "scr.gfx.pad", &game_theme.pad, 0 },
 	{ "scr.gfx.x", &game_theme.gfx_x, TF_POSX },
 	{ "scr.gfx.y", &game_theme.gfx_y, TF_POSY },
 	{ "scr.gfx.w", &game_theme.max_scene_w, TF_NEG },
@@ -539,7 +539,7 @@ int theme_img_scale(img_t *p)
 	pic = gfx_scale(*p, v, v, SCALABLE_THEME_SMOOTH);
 	if (!pic)
 		return -1;
-	gfx_free_image(*p); 
+	gfx_free_image(*p);
 	*p = pic;
 	return 0;
 }
@@ -563,7 +563,7 @@ static  int game_theme_scale(int w, int h)
 
 	xs = (float)w / (float)t->w;
 	ys = (float)h / (float)t->h;
-	
+
 	v = (xs < ys)?xs:ys;
 
 	if (!SCALABLE_THEME) {
@@ -727,18 +727,18 @@ static int theme_bg_scale(void)
 			xoff = 0;
 		if (yoff < 0)
 			yoff = 0;
-		if (t->scale != 1.0f || xoff || yoff) {
+		if (gfx_img_w(t->bg) < t->w || gfx_img_h(t->bg) < t->h) {
 			pic = gfx_new(t->w, t->h);
 			if (!pic)
 				return -1;
 			screen = gfx_screen(pic);
 			gfx_img_fill(pic, 0, 0, t->w, t->h, gfx_col(0,0,0));
-			gfx_draw(t->bg, xoff, yoff);
+			gfx_copy(t->bg, xoff, yoff);
 			gfx_screen(screen);
 			gfx_free_image(t->bg);
 			t->bg = pic;
 		}
-	}	
+	}
 	return 0;
 }
 
@@ -865,7 +865,7 @@ skip:
 	}
 
 	if (t->use_name && (t->changed & CHANGED_USE)) {
-		gfx_free_image(t->use);	
+		gfx_free_image(t->use);
 		if (!(t->use = gfx_load_image(t->use_name))) {
 			res = t->use_name;
 			goto err;
@@ -884,7 +884,7 @@ skip:
 			goto err;
 		gfx_set_cursor(t->cursor, t->cur_x, t->cur_y);
 	}
-	
+
 	if (t->menu_button_name && (t->changed & CHANGED_BUTTON)) {
 		gfx_free_image(t->menu_button);
 		if (!(t->menu_button = gfx_load_image(t->menu_button_name))) {
@@ -894,7 +894,7 @@ skip:
 		if (theme_img_scale(&t->menu_button))
 			goto err;
 	}
-	
+
 	if (t->click_name && (t->changed & CHANGED_CLICK)) {
 		sound_put(t->click);
 		t->click = sound_get(t->click_name);
@@ -918,9 +918,11 @@ err:
 
 int game_theme_update(void)
 {
-	game_release_theme();
+	game_theme_changed = 0;
+	game_release_theme(0);
 	if (game_theme_update_data()) {
 		fprintf(stderr, "Can not update theme!\n");
+		game_release_theme(1);
 		game_error();
 		return -1;
 	}
@@ -933,7 +935,7 @@ int game_theme_update(void)
 	return 0;
 }
 
-#if defined(ANDROID) || defined(IOS)
+#if defined(ANDROID) || defined(IOS) || defined(WINRT) || defined(_USE_SWROTATE)
 extern void rotate_landscape(void);
 extern void rotate_portrait(void);
 extern void unlock_rotation(void);
@@ -941,16 +943,15 @@ extern void unlock_rotation(void);
 
 int game_theme_init(void)
 {
-	color_t col = { .r = 0, .g = 0, .b = 0, .a = 0 };
 	int w  = opt_mode[0];
 	int h  = opt_mode[1];
 
 	game_cursor_show = 1;
 
 	if (opt_fs && opt_hires && !gfx_get_max_mode(&w, &h, MODE_ANY)) {
-#if defined(IOS) || defined(ANDROID)
+#if defined(IOS) || defined(ANDROID) || defined(WINRT) || defined(_USE_SWROTATE)
 		if ((game_theme.w > game_theme.h && w < h) ||
-			(game_theme.w < game_theme.h && w > h)) { /* rotated */
+		    (game_theme.w < game_theme.h && w > h)) { /* rotated */
 			if (gfx_get_max_mode(&w, &h, (game_theme.w > game_theme.h)?MODE_H:MODE_V)) {
 				gfx_get_max_mode(&w, &h, MODE_ANY); /* fallback to any mode */
 			}
@@ -963,34 +964,36 @@ int game_theme_init(void)
 			unlock_rotation();
 #endif
 	}
-#if defined(ANDROID) || defined(IOS)
+#if defined(ANDROID) || defined(IOS) || defined(WINRT) || defined(_USE_SWROTATE)
 	else {
 		unlock_rotation();
 	}
 #endif
 
 	if (w == -1) { /* as theme */
-#if !defined(IOS) /* IOS always hardware accelerated */
+#if !defined(IOS) && !defined(SAILFISHOS) /* always hardware accelerated */
 		if (gfx_get_max_mode(&w, &h, MODE_ANY) || (game_theme.w <= w && game_theme.h <= h)) {
 			w = opt_mode[0];
 			h = opt_mode[1];
 		}
 #endif
 	}
-
 	game_theme_scale(w, h);
-
 	if (gfx_set_mode(game_theme.w, game_theme.h, opt_fs)) {
 		opt_mode[0] = opt_mode[1] = -1; opt_fs = 0; /* safe options */
 		return -1;
 	}
-	gfx_fill(0, 0, game_theme.w, game_theme.h, col);
 	if (game_theme_update_data()) {
 		fprintf(stderr, "Can not init theme!\n");
 		game_theme_free();
 		game_theme_select(DEFAULT_THEME);
 		return -1;
 	}
+	gfx_bg(game_theme.bgcol);
+	if (!DIRECT_MODE)
+		game_clear_all();
+	gfx_flip();
+	gfx_commit();
 	return 0;
 }
 
@@ -1076,7 +1079,7 @@ static char *theme_name(const char *path, const char *d_name)
 		goto err;
 	l = lookup_lang_tag(p, "Name", ";", opt_lang);
 	free(p);
-	if (l) 
+	if (l)
 		return l;
 err:
 	return strdup(d_name);
@@ -1084,13 +1087,13 @@ err:
 
 static char *theme_name_idf(idf_t idf, const char *path, const char *d_name)
 {
-	char *l;	
+	char *l;
 	char *p = getfilepath(path, THEME_FILE);
 	if (!p)
 		goto err;
 	l = lookup_lang_tag_idf(idf, p, "Name", ";", opt_lang);
 	free(p);
-	if (l) 
+	if (l)
 		return l;
 err:
 	return strdup(d_name);
@@ -1296,7 +1299,7 @@ struct theme *theme_lookup(const char *name, int type)
 {
 	int i;
 	if (!name || !*name) {
-		if (themes_nr == 1 && themes[0].type == type) 
+		if (themes_nr == 1 && themes[0].type == type)
 			return &themes[0];
 		return NULL;
 	}
